**⬅️ 이전 시간: [Part 5.5: NumPy로 배우는 선형대수학](../05.5_linear_algebra_with_numpy/part_5.5_linear_algebra_with_numpy.md)**
**➡️ 다음 시간: [Part 6: 머신러닝 모델링과 평가](../06_machine_learning/part_6_machine_learning.md)**

---

# Part 5.6: 머신러닝/딥러닝을 위한 미적분

이전 파트에서 선형대수가 데이터와 모델을 표현하는 '언어'였다면, 미적분은 모델을 '학습'시키는 핵심 도구입니다. 특히 딥러닝에서 모델의 파라미터를 최적화하는 과정은 전적으로 미분의 원리에 기반합니다.

## 1. 미분 (Differentiation)

### 1.1. 미분의 개념

미분(differentiation)은 아주 짧은 순간의 **변화율(rate of change)** 또는 함수의 특정 지점에서의 **접선의 기울기(slope of the tangent line)**를 구하는 과정입니다.

함수 $y = f(x)$가 있을 때, $x$의 값이 아주 조금($\Delta x$) 변할 때 $y$의 값($\Delta y$)이 얼마나 변하는지를 나타내는 비율 $\frac{\Delta y}{\Delta x}$을 생각할 수 있습니다. 여기서 $\Delta x$를 거의 0에 가깝게 만들었을 때의 극한값이 바로 **도함수(derivative)**이며, $f'(x)$ 또는 $\frac{dy}{dx}$로 표기합니다.

$f'(x) = \lim_{\Delta x \to 0} \frac{f(x+\Delta x) - f(x)}{\Delta x}$

예를 들어, 함수 $f(x) = x^2$ 의 도함수는 $f'(x) = 2x$ 입니다. 이는 $x=1$에서의 접선의 기울기는 $2$, $x=2$에서의 접선의 기울기는 $4$라는 의미입니다.

### 1.2. 머신러닝에서의 미분의 역할: 비용 함수 최소화

머신러닝, 특히 딥러닝 모델의 학습 목표는 **비용 함수(Cost Function) 또는 손실 함수(Loss Function)의 값을 최소화**하는 것입니다. 비용 함수는 모델의 예측값과 실제 정답값의 차이(오차)를 나타내는 함수입니다.

가장 널리 사용되는 최적화 알고리즘인 **경사 하강법(Gradient Descent)**은 바로 이 미분 개념을 사용합니다.

1.  비용 함수를 모델의 파라미터(가중치 $W$, 편향 $b$)에 대해 편미분하여 **그래디언트(Gradient)**를 계산합니다. 그래디언트는 비용 함수가 가장 가파르게 증가하는 방향을 나타내는 벡터입니다.
2.  그래디언트의 **반대 방향**으로 파라미터를 조금씩 이동시킵니다. (가장 빠르게 감소하는 방향이므로)
3.  이 과정을 반복하면 비용 함수의 최솟값에 도달하게 되며, 이는 모델의 성능이 최적화되었음을 의미합니다.

> **그래디언트(Gradient)**: 다변수 함수에서 각 변수에 대한 편미분 값을 원소로 가지는 벡터. $\nabla f = (\frac{\partial f}{\partial x_1}, \frac{\partial f}{\partial x_2}, ..., \frac{\partial f}{\partial x_n})$

### 1.3. 기본 미분 공식

자주 사용되는 미분 공식은 다음과 같습니다.

*   **상수**: $f(x) = c \implies f'(x) = 0$
*   **거듭제곱**: $f(x) = x^n \implies f'(x) = nx^{n-1}$
*   **지수 함수**:
    *   $f(x) = e^x \implies f'(x) = e^x$
    *   $f(x) = a^x \implies f'(x) = a^x \ln(a)$
*   **로그 함수**: $f(x) = \ln(x) \implies f'(x) = \frac{1}{x}$
*   **시그모이드 함수**: $\sigma(x) = \frac{1}{1+e^{-x}} \implies \sigma'(x) = \sigma(x)(1-\sigma(x))$ (딥러닝에서 중요)

### 1.4. 연쇄 법칙 (Chain Rule)

연쇄 법칙은 **합성 함수(composite function)**를 미분하는 방법으로, 딥러닝의 역전파(Backpropagation) 알고리즘의 핵심 원리입니다.

두 함수 $y=f(u)$와 $u=g(x)$가 합성된 함수 $y = f(g(x))$의 도함수는 각 함수의 도함수의 곱으로 나타낼 수 있습니다.

$\frac{dy}{dx} = \frac{dy}{du} \cdot \frac{du}{dx}$

딥러닝 모델은 입력층부터 출력층까지 수많은 함수(레이어)가 연결된 거대한 합성 함수와 같습니다. 출력층에서 계산된 오차를 입력층 방향으로 다시 전파하며 각 레이어의 가중치에 대한 그래디언트를 계산할 때 이 연쇄 법칙이 사용됩니다.

## 2. Python을 이용한 미분

### 2.1. 기호적 미분 (Symbolic Differentiation) with `SymPy`

`SymPy`는 변수를 기호로 다루어 수학 공식을 그대로 사용하여 미분, 적분 등을 수행할 수 있는 라이브러리입니다.

```python
import sympy as sp

# 변수를 기호(symbol)로 선언
x = sp.Symbol('x')

# 함수 정의
f = x**3 + 2*x**2 + x + 5

# f를 x에 대해 미분
df_dx = sp.diff(f, x)

print(f"함수 f(x) = {f}")
print(f"도함수 f'(x) = {df_dx}")

# 특정 지점(x=2)에서의 미분 계수 계산
derivative_at_2 = df_dx.subs(x, 2)
print(f"f'(2) = {derivative_at_2}")
```

### 2.2. 수치 미분 (Numerical Differentiation) with `NumPy`

`NumPy`는 기호적 미분을 직접 지원하지 않지만, 미분의 정의를 이용하여 근사값을 계산하는 **수치 미분**을 구현할 수 있습니다.

$f'(x) \approx \frac{f(x+h) - f(x-h)}{2h}$ (중앙 차분)

여기서 $h$는 아주 작은 값(예: `1e-4`)입니다.

```python
import numpy as np

# 함수 정의
def f(x):
    return x**3 + 2*x**2 + x + 5

# 수치 미분 함수 구현
def numerical_derivative(func, x, h=1e-4):
    return (func(x + h) - func(x - h)) / (2 * h)

# x=2 에서의 미분 계수 계산
derivative_at_2 = numerical_derivative(f, 2)

print(f"f(x) = x^3 + 2x^2 + x + 5")
print(f"x=2 에서의 수치적 미분 계수: {derivative_at_2}")
# SymPy 결과(21)와 매우 유사한 값을 얻을 수 있습니다.
```

수치 미분은 구현이 간단하지만 근사값이기 때문에 오차가 발생할 수 있습니다. 반면 기호적 미분은 정확한 결과를 제공합니다. 딥러닝 프레임워크(TensorFlow, PyTorch)는 **자동 미분(Automatic Differentiation)**이라는 효율적이고 정확한 방식을 사용하여 그래디언트를 계산합니다. 자동 미분은 기호적 미분과 수치 미분의 장점을 결합한 방식입니다.

## 3. 적분 (Integration)

적분은 미분의 역연산으로, 크게 **부정적분(indefinite integral)**과 **정적분(definite integral)**으로 나뉩니다.

*   **부정적분**: 도함수가 주어졌을 때 원래 함수를 찾는 과정. $F'(x) = f(x)$일 때, $\int f(x)dx = F(x) + C$ (C는 적분 상수)
*   **정적분**: 함수 그래프의 특정 구간 아래의 **넓이**를 구하는 과정. $\int_{a}^{b} f(x)dx$

머신러닝에서 적분은 주로 확률 및 통계 분야에서 확률 분포의 기댓값이나 분산을 계산하는 등 이론적인 부분에서 활용됩니다.

## 4. 머신러닝/딥러닝 적용 실제 예시

### 4.1. 선형 회귀 모델에서의 경사 하강법

가장 간단한 머신러닝 모델 중 하나인 선형 회귀를 통해 그래디언트가 어떻게 계산되고 사용되는지 살펴보겠습니다.

1.  **모델 정의**: $y_{pred} = Wx + b$
    - $x$: 입력 데이터
    - $W, b$: 학습시킬 파라미터 (가중치와 편향)
    - $y_{pred}$: 모델의 예측값

2.  **비용 함수(Cost Function) 정의**: 평균 제곱 오차(Mean Squared Error, MSE)
    - $J(W, b) = \frac{1}{N} \sum_{i=1}^{N} (y_{true}^{(i)} - y_{pred}^{(i)})^2 = \frac{1}{N} \sum_{i=1}^{N} (y_{true}^{(i)} - (Wx^{(i)} + b))^2$
    - $y_{true}$: 실제 정답값

3.  **그래디언트 계산 (편미분)**
    - 비용 함수 $J$를 최소화하는 $W$와 $b$를 찾기 위해, 각각에 대해 편미분을 수행합니다.
    - **$W$에 대한 편미분 $\frac{\partial J}{\partial W}$**:
        - 연쇄 법칙에 따라, $J = \frac{1}{N}\sum u^2$, $u = y_{true} - v$, $v = Wx + b$ 로 볼 수 있습니다.
        - $\frac{\partial J}{\partial W} = \frac{\partial J}{\partial u} \frac{\partial u}{\partial v} \frac{\partial v}{\partial W}$
        - $\frac{\partial J}{\partial u} = \frac{1}{N}\sum 2u$, $\frac{\partial u}{\partial v} = -1$, $\frac{\partial v}{\partial W} = x$
        - 따라서, $\frac{\partial J}{\partial W} = \frac{1}{N} \sum 2(y_{true} - (Wx+b)) \cdot (-1) \cdot x = -\frac{2}{N} \sum x(y_{true} - y_{pred})$
    - **$b$에 대한 편미분 $\frac{\partial J}{\partial b}$**:
        - $\frac{\partial J}{\partial b} = \frac{\partial J}{\partial u} \frac{\partial u}{\partial v} \frac{\partial v}{\partial b}$
        - $\frac{\partial v}{\partial b} = 1$ 이므로,
        - $\frac{\partial J}{\partial b} = \frac{1}{N} \sum 2(y_{true} - (Wx+b)) \cdot (-1) \cdot 1 = -\frac{2}{N} \sum (y_{true} - y_{pred})$

4.  **파라미터 업데이트**
    - 계산된 그래디언트를 이용해 경사 하강법을 수행합니다. $\alpha$는 학습률(learning rate)입니다.
    - $W := W - \alpha \frac{\partial J}{\partial W}$
    - $b := b - \alpha \frac{\partial J}{\partial b}$

이 과정을 반복하면서 모델은 점차 오차를 줄여나가는 방향으로 $W$와 $b$를 업데이트하게 됩니다.

> **실제 구현 예시:**
> 이 과정은 Python과 `NumPy`를 사용하여 직접 구현해볼 수 있습니다. 다음 예제 코드를 통해 그래디언트 계산과 파라미터 업데이트가 실제로 어떻게 작동하는지 확인해보세요.
> - **코드:** [`linear_regression_gradient_descent.py`](./linear_regression_gradient_descent.py)

### 4.2. 간단한 신경망과 역전파 (Backpropagation)

더 복잡한 딥러닝 모델의 학습 원리인 역전파는 연쇄 법칙의 결정판입니다.

- **순전파(Forward Propagation)**: 입력 데이터가 신경망을 거쳐 최종 출력(예측)을 만들어내는 과정입니다.
- **역전파(Backward Propagation)**: 순전파 결과 계산된 오차(비용)를 출력층에서부터 입력층 방향으로 거꾸로 전파시키면서, 각 가중치 파라미터가 오차에 얼마나 기여했는지를(그래디언트) 연쇄 법칙을 이용해 계산하는 과정입니다.

예를 들어, `입력 -> 레이어1 -> 레이어2 -> 출력` 구조의 신경망에서 레이어1의 가중치에 대한 그래디언트는 다음과 같이 연쇄적으로 계산됩니다.

` (최종 오차 / 출력) -> (출력 / 레이어2 출력) -> (레이어2 출력 / 레이어1 출력) -> (레이어1 출력 / 레이어1 가중치) `

$\frac{\partial \text{Error}}{\partial W_1} = \frac{\partial \text{Error}}{\partial \text{Output}} \cdot \frac{\partial \text{Output}}{\partial \text{Layer2_out}} \cdot \frac{\partial \text{Layer2_out}}{\partial \text{Layer1_out}} \cdot \frac{\partial \text{Layer1_out}}{\partial W_1}$

이렇게 계산된 그래디언트($\frac{\partial \text{Error}}{\partial W_1}$)를 가지고 경사 하강법을 적용하여 레이어1의 가중치($W_1$)를 업데이트합니다. 이 과정을 모든 가중치에 대해 수행하는 것이 바로 딥러닝의 학습입니다.

---

이처럼 미적분, 특히 미분과 그래디언트의 개념은 머신러닝 모델이 데이터를 통해 '학습'하고 똑똑해지는 과정의 수학적 기반을 이룹니다. 

---

**⬅️ 이전 시간: [Part 5.5: NumPy로 배우는 선형대수학](../05.5_linear_algebra_with_numpy/part_5.5_linear_algebra_with_numpy.md)**
**➡️ 다음 시간: [Part 6: 머신러닝 모델링과 평가](../06_machine_learning/part_6_machine_learning.md)** 