import numpy as np
import matplotlib.pyplot as plt

# --- 1. 데이터 생성 ---
# y = 2x + 1 의 관계를 가지는 데이터를 생성합니다.
# 실제 데이터처럼 보이기 위해 약간의 노이즈를 추가합니다.
np.random.seed(42)
X = 2 * np.random.rand(100, 1)
y = 1 + 2 * X + np.random.randn(100, 1)

# --- 2. 모델 파라미터 초기화 및 하이퍼파라미터 설정 ---
# 학습을 통해 찾아야 할 가중치(W)와 편향(b)을 랜덤 값으로 초기화합니다.
W = np.random.randn(1, 1)
b = np.random.randn(1, 1)

# 학습률(learning_rate): 파라미터를 얼마나 업데이트할지 결정하는 값
learning_rate = 0.1
# 반복 학습 횟수(epochs)
epochs = 100

n_samples = len(X)
print(f"초기 W: {W[0][0]:.3f}, 초기 b: {b[0][0]:.3f}")

# --- 3. 경사 하강법 구현 ---
# 학습 과정에서 비용(cost)의 변화를 기록하기 위한 리스트
cost_history = []

for i in range(epochs):
    # 3-1. 예측값 계산 (Forward Pass)
    # y_pred = W * X + b
    y_pred = np.dot(X, W) + b

    # 3-2. 비용(Cost/Loss) 계산: 평균 제곱 오차(MSE)
    # cost = (1/N) * Σ(y_true - y_pred)^2
    cost = np.mean((y - y_pred)**2)
    cost_history.append(cost)
    
    # 3-3. 그래디언트 계산 (Backward Pass, 편미분)
    # dJ/dW = -(2/N) * Σ X * (y_true - y_pred)
    # dJ/db = -(2/N) * Σ (y_true - y_pred)
    # 문서의 수식과 비교해보세요. (여기서는 2를 곱하지 않고 평균 그래디언트를 사용)
    dW = (1/n_samples) * np.dot(X.T, (y_pred - y))
    db = (1/n_samples) * np.sum(y_pred - y)
    
    # 3-4. 파라미터 업데이트
    # W = W - learning_rate * dW
    # b = b - learning_rate * db
    W = W - learning_rate * dW
    b = b - learning_rate * db
    
    # 10번의 epoch마다 학습 결과 출력
    if (i + 1) % 10 == 0:
        print(f"Epoch {i+1:3d}: Cost={cost:.4f}, W={W[0][0]:.3f}, b={b[0][0]:.3f}")

print(f"\n학습 완료!")
print(f"최종 W: {W[0][0]:.3f}, 최종 b: {b[0][0]:.3f} (실제 값: W=2, b=1)")


# --- 4. 학습 결과 시각화 ---

# 1. 비용(Cost) 감소 그래프
plt.figure(figsize=(12, 5))
plt.subplot(1, 2, 1)
plt.plot(cost_history)
plt.title("Cost Reduction History")
plt.xlabel("Epoch")
plt.ylabel("Cost (MSE)")
plt.grid(True)

# 2. 회귀선 시각화
plt.subplot(1, 2, 2)
# 원본 데이터 산점도
plt.scatter(X, y, label='Original Data')
# 학습된 모델의 회귀선
plt.plot(X, np.dot(X, W) + b, color='red', linewidth=3, label='Fitted Line')
plt.title("Linear Regression Fit")
plt.xlabel("X")
plt.ylabel("y")
plt.legend()
plt.grid(True)

plt.tight_layout()
plt.show() 